/*
   Simple example of using a GLUT idle event handler
   to create an animation.
*/
#include <windows.h>
#include "glm.h"
#include "glext.h"  /* Needed for OpenGL Ver 1.2 features. */

// Prototypes for glut callback functions
void display( void );
void reshape(int, int);
void idle(void);

float animationAngle = 0.0;

int windowHeight;  // this is set by the glut reshape() callback function
int windowWidth;   // this is set by the glut reshape() callback function


GLMmodel* model = NULL;

#define rawImageWidth  256
#define rawImageHeight 256
GLubyte rawImage[rawImageHeight][rawImageWidth][3];

GLuint texName;

/*  Read a texture from a "raw" image file  */
void readRawImage(void)
{
   FILE* file;

   file = fopen("skin.raw", "rb");
   if (!file)
   {
      printf("Unable to open file: %s\n", "wood.raw");
      exit(1);
   }
   // Now read the raw image file into the array rawImage
   fread(rawImage, rawImageHeight * rawImageWidth * 3, 1, file);
   fclose(file);
}


int main(int argc, char** argv)
{
   // Standard GLUT initialization
   glutInit(&argc,argv);
   glutInitDisplayMode (GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
   glutInitWindowSize(500, 500);
   glutInitWindowPosition(150, 150);
   glutCreateWindow("Apple Model");
   glutDisplayFunc(display);
   glutReshapeFunc(reshape);
   glutIdleFunc(idle);

   // Initialize OpenGL state
   glClearColor(1.0, 1.0, 1.0, 1.0); // white background

   glEnable(GL_DEPTH_TEST);

   glShadeModel (GL_SMOOTH);

   glEnable(GL_LIGHTING);
   glEnable(GL_LIGHT0);
   /* Needs OpenGL Ver 1.2, which is not the default in Windows (OpenGL 1.1 is) */
   //glLightModeli(GL_LIGHT_MODEL_COLOR_CONTROL, GL_SEPARATE_SPECULAR_COLOR);

   readRawImage();
   glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

   glGenTextures(1, &texName);
   glBindTexture(GL_TEXTURE_2D, texName);

   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S,     GL_REPEAT);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T,     GL_REPEAT);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
   glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, rawImageWidth, rawImageHeight,
                               0, GL_RGB, GL_UNSIGNED_BYTE, rawImage);

   glEnable(GL_TEXTURE_2D);
   glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);  // try GL_DECAL
   glBindTexture(GL_TEXTURE_2D, texName);

   model = glmReadOBJ("apple.obj");

   glutMainLoop(); // Enter event loop
   return 0;       // We never get here
}//main()



void display()
{
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);  // clear the window

   // specify the viewport
   glViewport(0, 0, windowWidth, windowHeight); // left, bottom, width, height (in pixels)

   // set up viewing
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   glOrtho(-1.0, 1.0, -1.0, 1.0, -1.0, 1.0);

   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();

   //glColor3ub(255, 0, 0);

   glmUnitize(model);
   glPushMatrix();
     glRotatef(animationAngle, 0.0, 1.0, 1.0);
     glmFacetNormals(model);
     //glmVertexNormals(model, 90.0);
     //glmDraw(model, GLM_NONE );
     //glmDraw(model, GLM_FLAT );
     //glmDraw(model, GLM_SMOOTH );
     //glmDraw(model, GLM_MATERIAL );
     //glmDraw(model, GLM_TEXTURE );
     //glmDraw(model, GLM_MATERIAL | GLM_TEXTURE );
     //glmDraw(model, GLM_SMOOTH | GLM_MATERIAL );
     glmDraw(model, GLM_SMOOTH | GLM_MATERIAL | GLM_TEXTURE );
   glPopMatrix();

   glutSwapBuffers();
}//display()


void reshape(int width, int height)
{
   windowWidth  = width;
   windowHeight = height;

   printf("We got a \"reshape(%d, %d)\" event\n", width, height);

   glutPostRedisplay();
}//reshape()


void idle()
{
   animationAngle += 5.0;

   //printf("We got a \"idle()\" event, animationAngle = %f\n", animationAngle);

   //glutSetWindow(window);   // NOTE: You need this line if you are using a glui slider!!
   glutPostRedisplay();
}//idle()
