/*************************************************************
obj.c -- http://home.scarlet.be/zoetrope
Copyright (C) 2007  SFJOIBSUTUFZBFSU

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
MA  02110-1301, USA.
*************************************************************/
#include "obj.h"
#include <stdlib.h>
#include <string.h>

#define MAX_LINE	512

#ifndef MAX_PATH
	#define MAX_PATH	255
#endif

struct obj_err obj_error;

/*
 * Change the file name in path to filename.
 */
static char *change_filename(char *path, const char *filename) {
	char *p=strrchr(path,'\\');
	if (p==NULL)
		p=strrchr(path,'/');
	if (p)
		*(p+1)='\0';
	else
		path[0]='\0';
	strcat(path,filename);
	return path;
}

static char *read_line(char *buf, FILE *f) {
	size_t i;
	if (fgets(buf,MAX_LINE+1,f)==NULL)
		return NULL;
	for(i=0;i<MAX_LINE+1;i++) {
		if (buf[i]=='\n') {
			buf[i]='\0';
			return buf;
		}
	}
	return NULL;
}

static void strip_comment(char *buf) {
	char *p=strchr(buf,'#');
	if (p) *p='\0';
}

void free_mtl_list(struct obj_mtl *mtl) {
	struct obj_mtl *p;
	while(mtl) {
		p=mtl;
		mtl=mtl->next;
		free(p);
	}
}

static struct obj_mtl *load_mtl(struct obj_mtl **first_mtl, const char *filename) {
	FILE *f;
	char buf[MAX_LINE+1], stmt[MAX_LINE+1];
	struct obj_mtl *last_mtl=*first_mtl;
	while(last_mtl) {	/* go to last mtl in list */
		if (last_mtl->next!=NULL)
			last_mtl=last_mtl->next;
		else
			break;
	}
	strncpy(obj_error.filename,filename,FILENAME_MAX);
	if ((f=fopen(filename,"r"))==NULL) {
		obj_error.nerr=OBJ_EFOPEN;
		return NULL;
	}
	obj_error.nline=1;
	while(read_line(buf,f)) {
		strip_comment(buf);
		if (sscanf(buf,"%s",stmt)==1) {
			if (strcmp(stmt,"newmtl")==0) {
				char name[MAX_LINE+1];
				if (sscanf(buf,"%s %s",stmt,name)==2) {
					struct obj_mtl *mtl=(struct obj_mtl*)
										malloc(sizeof(struct obj_mtl));
					mtl->name=strdup(name);
					mtl->Ka[0]=0.2; mtl->Ka[1]=0.2; mtl->Ka[2]=0.2;
					mtl->Kd[0]=0.8; mtl->Kd[1]=0.8; mtl->Kd[2]=0.8;
					mtl->Ks[0]=1.0; mtl->Ks[1]=1.0; mtl->Ks[2]=1.0;
					mtl->Ns=1.0;
					mtl->map_Kd=NULL;
					mtl->map_Kd_id=0;
					mtl->next=NULL;
					if (last_mtl==NULL) {
						*first_mtl=last_mtl=mtl;
					} else {
						last_mtl->next=mtl;
						last_mtl=mtl;
					}
				} else {
					obj_error.nerr=OBJ_EBADNEWMTL;
					goto err;
				}
			} else if (strcmp(stmt,"Ka")==0) {
				if (last_mtl) {
					if (sscanf(buf,"%s %f %f %f",stmt,
						&last_mtl->Ka[0],
						&last_mtl->Ka[1],
						&last_mtl->Ka[2])==4) {
					} else {
						obj_error.nerr=OBJ_EBADKA;
						goto err;
					}
				} else {
					obj_error.nerr=OBJ_ENONEWMTL;
					goto err;
				}
			} else if (strcmp(stmt,"Kd")==0) {
				if (last_mtl) {
					if (sscanf(buf,"%s %f %f %f",stmt,
						&last_mtl->Kd[0],
						&last_mtl->Kd[1],
						&last_mtl->Kd[2])==4) {
					} else {
						obj_error.nerr=OBJ_EBADKD;
						goto err;
					}
				} else {
					obj_error.nerr=OBJ_ENONEWMTL;
					goto err;
				}
			} else if (strcmp(stmt,"Ks")==0) {
				if (last_mtl) {
					if (sscanf(buf,"%s %f %f %f",stmt,
						&last_mtl->Ks[0],
						&last_mtl->Ks[1],
						&last_mtl->Ks[2])==4) {
					} else {
						obj_error.nerr=OBJ_EBADKS;
						goto err;
					}
				} else {
					obj_error.nerr=OBJ_ENONEWMTL;
					goto err;
				}
			} else if (strcmp(stmt,"Ns")==0) {
				if (last_mtl) {
					if (sscanf(buf,"%s %f",stmt,&last_mtl->Ns)==2) {
						if (last_mtl->Ns<1.0 || last_mtl->Ns>128) {
							obj_error.nerr=OBJ_ENSRANGE;
							goto err;
						}
					} else {
						obj_error.nerr=OBJ_EBADNS;
						goto err;
					}
				} else {
					obj_error.nerr=OBJ_ENONEWMTL;
					goto err;
				}
			} else if (strcmp(stmt,"map_Kd")==0) {
				if (last_mtl) {
					char name[MAX_LINE+1], path[MAX_PATH+1];
					if (sscanf(buf,"%s %s",stmt,name)==2) {
						strcpy(path,filename);
						change_filename(path,name);
						last_mtl->map_Kd=strdup(path);
					} else {
						obj_error.nerr=OBJ_EBADMAPKD;
						goto err;
					}
				} else {
					obj_error.nerr=OBJ_ENONEWMTL;
					goto err;
				}
			}
		}
		obj_error.nline++;
	}
	fclose(f);
	return last_mtl;

err:
	fclose(f);
	return NULL;
}

/*
 * load_obj_model() will read the file twice: the first time to count the
 * number of vertices, texels and normals and the second time to fill in
 * the data structure. The file is read line by line.
 */
struct obj_model *load_obj_model(struct obj_model *m, const char *filename) {
	FILE *f;
	char buf[MAX_LINE+1], stmt[MAX_LINE+1];
	struct obj_node *last_node=NULL;
	strncpy(obj_error.filename,filename,FILENAME_MAX);
	if ((f=fopen(filename,"r"))==NULL) {
		free_obj_model(m);
		obj_error.nerr=OBJ_EFOPEN;
		return NULL;
	}
	memset(m,0,sizeof(struct obj_model));
	obj_error.nline=1;
	while(read_line(buf,f)) {
		if (sscanf(buf,"%s",stmt)==1) {
			if (strcmp(stmt,"v")==0)
				m->nv++;
			else if (strcmp(stmt,"vt")==0)
				m->nvt++;
			else if (strcmp(stmt,"vn")==0)
				m->nvn++;
		}
	}
	m->v=(struct obj_v*)malloc(sizeof(struct obj_v)*m->nv);
	m->vt=(struct obj_vt*)malloc(sizeof(struct obj_vt)*m->nvt);
	m->vn=(struct obj_vn*)malloc(sizeof(struct obj_vn)*m->nvn);
	m->nv=0;
	m->nvt=0;
	m->nvn=0;
	fseek(f,0,SEEK_SET);
	while(read_line(buf,f)) {
		strip_comment(buf);
		if (sscanf(buf,"%s",stmt)==1) {
			if (strcmp(stmt,"v")==0) {
				float x, y, z;
				if (sscanf(buf,"%s %f %f %f",stmt,&x,&y,&z)==4) {
					m->v[m->nv].x=x;
					m->v[m->nv].y=y;
					m->v[m->nv].z=z;
					m->nv++;
				} else {
					obj_error.nerr=OBJ_EBADV;
					goto err;
				}
			} else if (strcmp(stmt,"vt")==0) {
				float u, v;
				int i=sscanf(buf,"%s %f %f",stmt,&u,&v);
				if (i==2) {
					m->vt[m->nvt].u=u;
					m->vt[m->nvt].v=0;
				} else if (i==3) {
					m->vt[m->nvt].u=u;
					m->vt[m->nvt].v=v;
				} else {
					obj_error.nerr=OBJ_EBADVT;
					goto err;
				}
				m->nvt++;
			} else if (strcmp(stmt,"vn")==0) {
				float x, y, z;
				if (sscanf(buf,"%s %f %f %f",stmt,&x,&y,&z)==4) {
					m->vn[m->nvn].x=x;
					m->vn[m->nvn].y=y;
					m->vn[m->nvn].z=z;
					m->nvn++;
				} else {
					obj_error.nerr=OBJ_EBADVN;
					goto err;
				}
			} else if (strcmp(stmt,"f")==0) {
				char s[MAX_LINE+1], *p;
				int fmt=-2, i, j, k, n;
				struct obj_f *pf;
				struct obj_node *node;
				pf=(struct obj_f*)malloc(sizeof(struct obj_f));
				pf->n=0;
				strcpy(s,buf);			/* strtok() puts '\0' in buf */
				strtok(s," \r\n\t");
				while (p=strtok(NULL," \r\n\t")) {
					pf->n++;
					if (fmt==-2) {
						if (sscanf(p,"%d//%d",&j,&j)==2) {
							fmt=0;
						} else {
							i=sscanf(p,"%d/%d/%d",&j,&j,&j);
							switch(i) {
								case 3: fmt=3; break;
								case 1: fmt=1; break;
								case EOF:
									free(pf);
									obj_error.nerr=OBJ_EBADF;
									goto err;
							}
						}
					}
				}
				if (pf->n<3) {
					free(pf);
					obj_error.nerr=OBJ_EBADF;
					goto err;
				}
				strtok(buf," \r\n\t");
				if (fmt==0) {
					pf->v=(int*)malloc(sizeof(int)*pf->n);
					pf->vt=NULL;
					pf->vn=(int*)malloc(sizeof(int)*pf->n);
					n=0;
					while (p=strtok(NULL," \r\n\t")) {
						if (sscanf(p,"%d//%d",&i,&k)==2) {
							if (i>=0)
								pf->v[n]=i-1;
							else
								pf->v[n]=m->nv+i;
							if (k>=0)
								pf->vn[n]=k-1;
							else
								pf->vn[n]=m->nvn+k;
						} else {
							free(pf);
							obj_error.nerr=OBJ_EBADF;
							goto err;
						}
						n++;
					}
				} else if (fmt==1) {
					pf->v=(int*)malloc(sizeof(int)*pf->n);
					pf->vt=NULL;
					pf->vn=NULL;
					n=0;
					while (p=strtok(NULL," \r\n\t")) {
						if (sscanf(p,"%d",&i)==1) {
							if (i>=0)
								pf->v[n]=i-1;
							else
								pf->v[n]=m->nv+i;
						} else {
							free(pf);
							obj_error.nerr=OBJ_EBADF;
							goto err;
						}
						n++;
					}
				} else if (fmt==3) {
					pf->v=(int*)malloc(sizeof(int)*pf->n);
					pf->vt=(int*)malloc(sizeof(int)*pf->n);
					pf->vn=(int*)malloc(sizeof(int)*pf->n);
					n=0;
					while (p=strtok(NULL," \r\n\t")) {
						if (sscanf(p,"%d/%d/%d",&i,&j,&k)==3) {
							if (i>=0)
								pf->v[n]=i-1;
							else
								pf->v[n]=m->nv+i;
							if (j>=0)
								pf->vt[n]=j-1;
							else
								pf->vt[n]=m->nvt+j;
							if (k>=0)
								pf->vn[n]=k-1;
							else
								pf->vn[n]=m->nvn+k;
						} else {
							free(pf);
							obj_error.nerr=OBJ_EBADF;
							goto err;
						}
						n++;
					}
				}
				node=(struct obj_node*)malloc(sizeof(struct obj_node));
				node->id=OBJ_NODE_F;
				node->data=pf;
				node->next=NULL;
				if (last_node==NULL) {
					m->first_node=last_node=node;
				} else {
					last_node->next=node;
					last_node=node;
				}
			} else if (strcmp(stmt,"mtllib")==0) {
				char s[MAX_LINE+1], *p, *q, path[MAX_PATH+1];
				int nline=obj_error.nline, n;
				strcpy(s,buf);			/* strtok() puts '\0' in buf */
				strtok(s," \r\n\t");
				n=0;
				while (p=strtok(NULL," \r\n\t")) {
					strcpy(path,filename);
					change_filename(path,p);
					if (load_mtl(&m->first_mtl,path)==NULL)
						goto err;
					n++;
				}
				strncpy(obj_error.filename,filename,FILENAME_MAX);
				obj_error.nline=nline;
				if (n==0) {
					obj_error.nerr=OBJ_EBADMTLLIB;
					goto err;
				}
			} else if (strcmp(stmt,"usemtl")==0) {
				char name[MAX_LINE+1];
				if (sscanf(buf,"%s %s",stmt,name)==2) {
					struct obj_mtl *mtl=m->first_mtl;
					while(mtl) {
						if (strcmp(mtl->name,name)==0) {
							struct obj_node *node;
							node=(struct obj_node*)malloc(sizeof(struct obj_node));
							node->id=OBJ_NODE_USEMTL;
							node->data=mtl;
							node->next=NULL;
							if (last_node==NULL) {
								m->first_node=last_node=node;
							} else {
								last_node->next=node;
								last_node=node;
							}
							break;
						}
						mtl=mtl->next;
					}
					if (mtl==NULL) {
						obj_error.nerr=OBJ_ENOUSEMTL;
						goto err;
					}
				} else {
					obj_error.nerr=OBJ_EBADUSEMTL;
					goto err;
				}
			}
		}
		obj_error.nline++;
	}
	fclose(f);
	return m;

err:
	fclose(f);
	free_obj_model(m);
	return NULL;
}

void free_obj_model(struct obj_model *m) {
	struct obj_node *node;
	struct obj_f *pf;
	struct obj_mtl *pmtl;
	free(m->v);
	free(m->vt);
	free(m->vn);
	while(m->first_node) {
		switch(m->first_node->id) {
			case OBJ_NODE_F:
				pf=(struct obj_f*)m->first_node->data;
				free(pf->v);
				free(pf->vt);
				free(pf->vn);
				break;
			case OBJ_NODE_USEMTL:
				pmtl=(struct obj_mtl*)m->first_node->data;
				if (pmtl->map_Kd)
					free(pmtl->map_Kd);
				break;
		}
		free(m->first_node->data);
		node=m->first_node;
		m->first_node=m->first_node->next;
		free(node);
	}
	free_mtl_list(m->first_mtl);
}

void obj_perror(const char *s) {
	if (s && strlen(s)>0)
		printf("%s: ",s);
	switch(obj_error.nerr) {
		case OBJ_EFOPEN:
			printf("cannot open file '%s'\n",obj_error.filename);
			break;
		case OBJ_EBADV:
			printf("bad 'v' statement at line %d in file '%s'\n",
				obj_error.nline,obj_error.filename);
			break;
		case OBJ_EBADVT:
			printf("bad 'vt' statement at line %d in file '%s'\n",
				obj_error.nline,obj_error.filename);
			break;
		case OBJ_EBADVN:
			printf("bad 'vn' statement at line %d in file '%s'\n",
				obj_error.nline,obj_error.filename);
			break;
		case OBJ_EBADF:
			printf("bad 'f' statement at line %d in file '%s'\n",
				obj_error.nline,obj_error.filename);
			break;
		case OBJ_EBADNEWMTL:
			printf("bad 'newmtl' statement at line %d in file '%s'\n",
				obj_error.nline,obj_error.filename);
			break;
		case OBJ_EBADKA:
			printf("bad 'Ka' statement at line %d in file '%s'\n",
				obj_error.nline,obj_error.filename);
			break;
		case OBJ_EBADKD:
			printf("bad 'Kd' statement at line %d in file '%s'\n",
				obj_error.nline,obj_error.filename);
			break;
		case OBJ_EBADKS:
			printf("bad 'Ks' statement at line %d in file '%s'\n",
				obj_error.nline,obj_error.filename);
			break;
		case OBJ_ENSRANGE:
			printf("'Ns' out of range at line %d in file '%s'\n",
				obj_error.nline,obj_error.filename);
			break;
		case OBJ_EBADNS:
			printf("bad 'Ns' statement at line %d in file '%s'\n",
				obj_error.nline,obj_error.filename);
			break;
		case OBJ_EBADMAPKD:
			printf("bad 'map_Kd' statement at line %d in file '%s'\n",
				obj_error.nline,obj_error.filename);
			break;
		case OBJ_ENONEWMTL:
			printf("missing 'newmtl' statement at line %d in file '%s'\n",
				obj_error.nline,obj_error.filename);
			break;
		case OBJ_EBADUSEMTL:
			printf("bad 'usemtl' statement at line %d in file '%s'\n",
				obj_error.nline,obj_error.filename);
			break;
		case OBJ_EBADMTLLIB:
			printf("bad 'mtllib' statement at line %d in file '%s'\n",
				obj_error.nline,obj_error.filename);
			break;
		case OBJ_ENOUSEMTL:
			printf("unknown material name at line %d in file '%s'\n",
				obj_error.nline,obj_error.filename);
			break;
		default:
			printf("\n");
			break;
	}
}

/*
 * get_obj_bounding_box() returns the bounding box for the model in
 * minmax (minx,maxx,miny,maxy,minz,maxz).
 */
void get_obj_bounding_box(const struct obj_model *m, float *minmax) {
	int i;
	memset(minmax,0,sizeof(float)*6);
	float *minx=&minmax[0];
	float *maxx=&minmax[1];
	float *miny=&minmax[2];
	float *maxy=&minmax[3];
	float *minz=&minmax[4];
	float *maxz=&minmax[5];
	for(i=0;i<m->nv;i++) {
		if (m->v[i].x<*minx) *minx=m->v[i].x;
		if (m->v[i].x>*maxx) *maxx=m->v[i].x;
		if (m->v[i].y<*miny) *miny=m->v[i].y;
		if (m->v[i].y>*maxy) *maxy=m->v[i].y;
		if (m->v[i].z<*minz) *minz=m->v[i].z;
		if (m->v[i].z>*maxz) *maxz=m->v[i].z;
	}
}

