/*************************************************************
viewobj -- http://home.scarlet.be/zoetrope
Copyright (C) 2007  SFJOIBSUTUFZBFSU

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
MA  02110-1301, USA.
*************************************************************/
#include <gl\freeglut.h>
#include <gl\glext.h>
#include <math.h>
#include "obj.h"

/*
 * gluPerspective stuff.
 */
#define FOVY			30.0
#define ZNEAR			1.0
#define ZFAR			100.0

/*
 * ViewObj will automatically center the object and
 * scale it to AUTOFOCUS_SIZE so that it fits nicely
 * in the window. See auto_focus().
 */
#define AUTOFOCUS_SIZE	10.0

/*
 * Command line processing stuff.
 */
#define OPT_HELP		1
#define OPT_INFO		2

/*
 * Distance between lights.
 */
#define LIGHT_SPACING	100.0

struct layer {
	int width, height;
	unsigned char *rgba;
};

int prevx=-1, prevy=-1;
GLfloat rotx=0, roty=0, rotz=0;
GLfloat transx=0, transy=0, transz=0;
GLfloat scale=1.0;
GLfloat camera_z=30.0;
struct obj_model model;
GLenum polygon_mode=GL_LINE;
GLenum shade_model=GL_FLAT;

/*
 * Draw an obj_model. This is so straightforward that it
 * doesn't need any explanation.
 */
static void draw_obj_model(const struct obj_model *m) {
	struct obj_node *node;
	struct obj_f *f;
	struct obj_v *v;
	struct obj_vt *vt;
	struct obj_vn *vn;
	struct obj_mtl *mtl;
	int i;
	node=m->first_node;
	while(node) {
		if (node->id==OBJ_NODE_F) {
			f=(struct obj_f*)node->data;
			glBegin(GL_POLYGON);
			for(i=0;i<f->n;i++) {
				v=&m->v[f->v[i]];
				if (f->vn) {
					vn=&m->vn[f->vn[i]];
					glNormal3f(vn->x,vn->y,vn->z);
				}
				if (f->vt) {
					vt=&m->vt[f->vt[i]];
					glTexCoord2f(vt->u,vt->v);
				}
				glVertex3f(v->x,v->y,v->z);
			}
			glEnd();
		} else if (node->id==OBJ_NODE_USEMTL) {
			mtl=(struct obj_mtl*)node->data;
			GLfloat a[4];
			a[0]=mtl->Ka[0];
			a[1]=mtl->Ka[1];
			a[2]=mtl->Ka[2];
			a[3]=1.0;
			glMaterialfv(GL_FRONT,GL_AMBIENT,a);
			a[0]=mtl->Kd[0];
			a[1]=mtl->Kd[1];
			a[2]=mtl->Kd[2];
			a[3]=1.0;
			glMaterialfv(GL_FRONT,GL_DIFFUSE,a);
			a[0]=mtl->Ks[0];
			a[1]=mtl->Ks[1];
			a[2]=mtl->Ks[2];
			a[3]=1.0;
			glMaterialfv(GL_FRONT,GL_SPECULAR,a);
			glMaterialf(GL_FRONT,GL_SHININESS,mtl->Ns);
			if (mtl->map_Kd_id!=0) {
				glBindTexture(GL_TEXTURE_2D,mtl->map_Kd_id);
			}
		}
		node=node->next;
	}
}

static void display(void) {
	glColor4f(1,1,1,1);
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	gluLookAt(0,0,camera_z,0,0,0,0,1,0);
	glRotatef(roty,0,1,0);	/* left right */
	glRotatef(rotx,1,0,0);	/* up down */
	glScalef(scale,scale,scale);
	glTranslatef(transx,transy,transz);
	draw_obj_model(&model);
	glFlush();
	glutSwapBuffers();
}

static void mouse(int button, int state, int x, int y) {
	if (button==GLUT_LEFT_BUTTON) {
		if (state==GLUT_DOWN) {
			prevx=x;
			prevy=y;
		} else if (state==GLUT_UP) {
			prevx=-1;
			prevy=-1;
		}
	}
}

static void motion(int x, int y) {
	if (prevx!=-1) {
		roty+=x-prevx;
		if (roty>=360) roty-=360;
		rotx+=y-prevy;
		if (rotx>=360) rotx-=360;
		glutPostRedisplay();
		prevx=x;
		prevy=y;
	}
}

static void reshape(int width, int height) {
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(FOVY,(double)width/height,ZNEAR,ZFAR);
	glViewport(0,0,width,height);
}

static void keyboard(unsigned char key, int x, int y) {
	GLboolean b;
	GLint i[2];
	switch(key) {
		case 'm':
			glGetIntegerv(GL_POLYGON_MODE,i);
			if (i[0]==GL_LINE)
				glPolygonMode(GL_FRONT,GL_FILL);
			else
				glPolygonMode(GL_FRONT,GL_LINE);
			break;
		case 's':
			glGetIntegerv(GL_SHADE_MODEL,i);
			if (i[0]==GL_FLAT)
				glShadeModel(GL_SMOOTH);
			else
				glShadeModel(GL_FLAT);
			break;
		case 't':
			glGetBooleanv(GL_TEXTURE_2D,&b);
			if (b)
				glDisable(GL_TEXTURE_2D);
			else
				glEnable(GL_TEXTURE_2D);
			break;
		case 'l':
			glGetBooleanv(GL_LIGHTING,&b);
			if (b)
				glDisable(GL_LIGHTING);
			else
				glEnable(GL_LIGHTING);
			break;
		case '+':
			camera_z-=1.0;
			break;
		case '-':
			camera_z+=1.0;
			break;
		case '1':
			glGetBooleanv(GL_LIGHT0,&b);
			if (b)
				glDisable(GL_LIGHT0);
			else
				glEnable(GL_LIGHT0);
			break;
		case '2':
			glGetBooleanv(GL_LIGHT1,&b);
			if (b)
				glDisable(GL_LIGHT1);
			else
				glEnable(GL_LIGHT1);
			break;
		case '3':
			glGetBooleanv(GL_LIGHT2,&b);
			if (b)
				glDisable(GL_LIGHT2);
			else
				glEnable(GL_LIGHT2);
			break;
		case '4':
			glGetBooleanv(GL_LIGHT3,&b);
			if (b)
				glDisable(GL_LIGHT3);
			else
				glEnable(GL_LIGHT3);
			break;
		case '6':
			glGetBooleanv(GL_LIGHT4,&b);
			if (b)
				glDisable(GL_LIGHT4);
			else
				glEnable(GL_LIGHT4);
			break;
		case '7':
			glGetBooleanv(GL_LIGHT5,&b);
			if (b)
				glDisable(GL_LIGHT5);
			else
				glEnable(GL_LIGHT5);
			break;
		case '8':
			glGetBooleanv(GL_LIGHT6,&b);
			if (b)
				glDisable(GL_LIGHT6);
			else
				glEnable(GL_LIGHT6);
			break;
		case '9':
			glGetBooleanv(GL_LIGHT7,&b);
			if (b)
				glDisable(GL_LIGHT7);
			else
				glEnable(GL_LIGHT7);
			break;
	}
	glutPostRedisplay();
}

static int pack(const unsigned char *buf) {
	int n=buf[0];
	n<<=8;
	n|=buf[1];
	n<<=8;
	n|=buf[2];
	n<<=8;
	n|=buf[3];
	return n;
}

static int read_layer(struct layer *layer, FILE *f) {
	unsigned char buf[4];
	int i, j;
	if (fread(buf,sizeof(unsigned char),4,f)<4) return 0;
	layer->width=pack(buf);
	if (fread(buf,sizeof(unsigned char),4,f)<4) return 0;
	layer->height=pack(buf);
	layer->rgba=(unsigned char*)malloc(sizeof(unsigned char)*
		layer->width*layer->height*4);
	if (fread(layer->rgba,sizeof(unsigned char),layer->width*
		layer->height*4,f)<layer->width*layer->height*4) {
		free(layer->rgba);
		return 0;
	}

	/*
	 * Flip image. glTexImage2D() wants the first element to
	 * correspond to the lower left corner of the texture image.
	 */

	for(i=0;i<layer->height/2;i++) {
		for(j=0;j<layer->width;j++) {
			int offset1=(i*layer->width+j)*4;
			int offset2=(layer->width*layer->height*4)-
						((i+1)*layer->width*4)+(j*4);
			memcpy(buf,&layer->rgba[offset1],4*sizeof(unsigned char));
			memcpy(&layer->rgba[offset1],&layer->rgba[offset2],
				4*sizeof(unsigned char));
			memcpy(&layer->rgba[offset2],buf,4*sizeof(unsigned char));
		}
	}
	return 1;
}

/*
 * preload_textures() will try to load the file map_Kd (see
 * struct obj_mtl), create an OpenGL texture and assign the
 * texture id to map_Kd_id. This way, load_obj_model() is
 * completely independent of OpenGL.
 */
static void preload_textures(struct obj_model *m) {
	struct obj_node *node;
	node=m->first_node;
	while(node) {
		if (node->id==OBJ_NODE_USEMTL) {
			struct obj_mtl *p=(struct obj_mtl*)node->data;
			if (p->map_Kd) {
				struct layer l;
				FILE *f;
				if ((f=fopen(p->map_Kd,"rb"))!=NULL) {
					if (read_layer(&l,f)) {
						glGenTextures(1,&p->map_Kd_id);
						glBindTexture(GL_TEXTURE_2D,p->map_Kd_id);
						glTexImage2D(GL_TEXTURE_2D,0,4,l.width,l.height,
							0,GL_RGBA,GL_UNSIGNED_BYTE,l.rgba);
						glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
						glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
						glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
						glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
					} else {
						printf("cannot read layer '%s'\n",p->map_Kd);
					}
				} else {
					printf("cannot open file '%s'\n",p->map_Kd);
				}
			}
		}
		node=node->next;
	}
}

static void delete_textures(struct obj_model *m) {
	struct obj_node *node;
	node=m->first_node;
	while(node) {
		if (node->id==OBJ_NODE_USEMTL) {
			struct obj_mtl *p=(struct obj_mtl*)node->data;
			if (p->map_Kd_id>0) {
				glDeleteTextures(1,&p->map_Kd_id);
			}
		}
		node=node->next;
	}
}

/*
 * auto_focus() will automatically center the object and
 * scale it to AUTOFOCUS_SIZE so that it fits nicely in
 * our window. get_obj_bounding_box() will return the
 * min and max values of x, y, and z of model m in bbox.
 */
static void auto_focus(const struct obj_model *m) {
	float bbox[6], dx, dy, dz;
	get_obj_bounding_box(m,bbox);
	dx=bbox[1]-bbox[0];
	dy=bbox[3]-bbox[2];
	dz=bbox[5]-bbox[4];
	transx=0.0-(bbox[0]+dx/2.0);
	transy=0.0-(bbox[2]+dy/2.0);
	transz=0.0-(bbox[4]+dz/2.0);
	scale=dx;
	if (dy>scale) scale=dy;
	if (dz>scale) scale=dz;
	scale=AUTOFOCUS_SIZE/scale;
}

/*
 * Return the filename in path.
 */
char *fname(char *path) {
	char *p=strrchr(path,'\\');
	if (p==NULL)
		p=strrchr(path,'/');
	if (p)
		p++;
	else
		p=path;
	return p;
}

void usage(const char *filename) {
	printf("Usage: %s [options] file\n",filename);
	puts("Options:");
	puts("  -?\tDisplay this information");
	puts("  -i\tDisplay file information");
}

int main(int argc, char **argv) {

	char *filename=NULL, title[FILENAME_MAX+1];
	int i, opt=0;

	/*
	 * 8 lights. light_position[0] is in the lower left corner:
	 *
	 * 		5-6-7
	 * 		| | |		
	 * 		3-*-4
	 * 		| | |
	 * 		0-1-2
	 */
	GLfloat light_position[9][4]={
		{-LIGHT_SPACING,-LIGHT_SPACING,LIGHT_SPACING,1.0},
		{0.0,-LIGHT_SPACING,LIGHT_SPACING,1.0},
		{LIGHT_SPACING,-LIGHT_SPACING,LIGHT_SPACING,1.0},
		{-LIGHT_SPACING,0.0,LIGHT_SPACING,1.0},
		/*{0.0,0.0,LIGHT_SPACING,1.0},*/
		{LIGHT_SPACING,0.0,LIGHT_SPACING,1.0},
		{-LIGHT_SPACING,LIGHT_SPACING,LIGHT_SPACING,1.0},
		{0.0,LIGHT_SPACING,LIGHT_SPACING,1.0},
		{LIGHT_SPACING,LIGHT_SPACING,LIGHT_SPACING,1.0}
	};

	GLfloat light_ambient[]={ 0.0, 0.0, 0.0, 1.0 };
	GLfloat light_diffuse[]={ 1.0, 1.0, 1.0, 1.0 };
	GLfloat light_specular[]={ 1.0, 1.0, 1.0, 1.0 };

	for (i=1;i<argc;i++) {
		if (argv[i][0]=='-') {
			if (argv[i][1]=='?')
				opt=OPT_HELP;
			else if (argv[i][1]=='i')
				opt=OPT_INFO;
		} else {
			if (filename==NULL) {
				filename=argv[i];
			}
		}
	}

	if (opt==OPT_HELP) {
		usage(fname(argv[0]));
		return 0;
	}

	if (filename==NULL) {
		printf("%s: no input file\n",fname(argv[0]));
		return 0;
	}

	if (load_obj_model(&model,filename)==NULL) {
		obj_perror(NULL);
		return -1;
	}

	if (opt==OPT_INFO) {
		printf("%d vertices\n",model.nv);
		printf("%d texels\n",model.nvt);
		printf("%d normals\n",model.nvn);
		return 0;
	}

	strcpy(title,"ViewObj - ");
	strcat(title,filename);

	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGBA|GLUT_DOUBLE|GLUT_DEPTH);
	glutSetOption(GLUT_ACTION_ON_WINDOW_CLOSE,GLUT_ACTION_CONTINUE_EXECUTION);
	glutCreateWindow(title);
	glutMouseFunc(mouse);
	glutMotionFunc(motion);
	glutReshapeFunc(reshape);
	glutDisplayFunc(display);
	glutKeyboardFunc(keyboard);

	preload_textures(&model);

	auto_focus(&model);

	glClearColor(0,0,0,1);

	glShadeModel(GL_FLAT);
	glPolygonMode(GL_FRONT, GL_LINE);

	glEnable(GL_CULL_FACE);
	glEnable(GL_DEPTH_TEST);

	/*
	 * GL_MODULATE multiplies the primitive's lighting color components
	 * with the texture color components. The call to glLightModeli()
	 * makes specular highlighting work.
	 */

	glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
	glLightModeli(GL_LIGHT_MODEL_COLOR_CONTROL,GL_SEPARATE_SPECULAR_COLOR);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT,GL_FASTEST);

	/*
	 * Lights.
	 */

	glLightfv(GL_LIGHT0,GL_POSITION,light_position[0]);
	glLightfv(GL_LIGHT0,GL_AMBIENT,light_ambient);
	glLightfv(GL_LIGHT0,GL_DIFFUSE,light_diffuse);
	glLightfv(GL_LIGHT0,GL_SPECULAR,light_specular);
	glEnable(GL_LIGHT0);

	glLightfv(GL_LIGHT1,GL_POSITION,light_position[1]);
	glLightfv(GL_LIGHT1,GL_AMBIENT,light_ambient);
	glLightfv(GL_LIGHT1,GL_DIFFUSE,light_diffuse);
	glLightfv(GL_LIGHT1,GL_SPECULAR,light_specular);

	glLightfv(GL_LIGHT2,GL_POSITION,light_position[2]);
	glLightfv(GL_LIGHT2,GL_AMBIENT,light_ambient);
	glLightfv(GL_LIGHT2,GL_DIFFUSE,light_diffuse);
	glLightfv(GL_LIGHT2,GL_SPECULAR,light_specular);

	glLightfv(GL_LIGHT3,GL_POSITION,light_position[3]);
	glLightfv(GL_LIGHT3,GL_AMBIENT,light_ambient);
	glLightfv(GL_LIGHT3,GL_DIFFUSE,light_diffuse);
	glLightfv(GL_LIGHT3,GL_SPECULAR,light_specular);

	glLightfv(GL_LIGHT4,GL_POSITION,light_position[4]);
	glLightfv(GL_LIGHT4,GL_AMBIENT,light_ambient);
	glLightfv(GL_LIGHT4,GL_DIFFUSE,light_diffuse);
	glLightfv(GL_LIGHT4,GL_SPECULAR,light_specular);

	glLightfv(GL_LIGHT5,GL_POSITION,light_position[5]);
	glLightfv(GL_LIGHT5,GL_AMBIENT,light_ambient);
	glLightfv(GL_LIGHT5,GL_DIFFUSE,light_diffuse);
	glLightfv(GL_LIGHT5,GL_SPECULAR,light_specular);

	glLightfv(GL_LIGHT6,GL_POSITION,light_position[6]);
	glLightfv(GL_LIGHT6,GL_AMBIENT,light_ambient);
	glLightfv(GL_LIGHT6,GL_DIFFUSE,light_diffuse);
	glLightfv(GL_LIGHT6,GL_SPECULAR,light_specular);

	glLightfv(GL_LIGHT7,GL_POSITION,light_position[7]);
	glLightfv(GL_LIGHT7,GL_AMBIENT,light_ambient);
	glLightfv(GL_LIGHT7,GL_DIFFUSE,light_diffuse);
	glLightfv(GL_LIGHT7,GL_SPECULAR,light_specular);

	glutMainLoop();

	delete_textures(&model);

	return 0;

}

